import copy
from unittest import TestCase
from typing import Type, cast
from dataclasses import dataclass, field

from kube_models import get_k8s_resource_model, K8sResource, Loadable
from kube_models.api_v1.io.k8s.apimachinery.pkg.apis.meta.v1 import ObjectMeta
from kube_models.api_v1.io.k8s.api.core.v1 import Secret, Namespace
from kube_models.apis_apps_v1.io.k8s.api.apps.v1 import Deployment


class UtilsTest(TestCase):
    def test_model_by_kind_core(self):
        secret = cast(Type[Secret], get_k8s_resource_model('v1', 'Secret'))
        self.assertEqual("v1", secret.apiVersion)
        self.assertEqual("Secret", secret.kind)
        self.assertIsNone(None, secret.group_)
        self.assertEqual("api/v1/namespaces/{namespace}/secrets", secret.api_path())
        self.assertEqual("secrets", secret.plural_)
        self.assertEqual(True, Secret.is_namespaced_)
        self.assertEqual(False, Namespace.is_namespaced_)

    def test_model_by_kind_apis_group(self):
        deployment = cast(Type[Deployment], get_k8s_resource_model("apps/v1", "Deployment"))
        self.assertEqual("apps/v1", deployment.apiVersion)
        self.assertEqual("Deployment", deployment.kind)
        self.assertEqual("apps", deployment.group_)
        self.assertEqual("apis/apps/v1/namespaces/{namespace}/deployments", deployment.api_path())
        self.assertEqual("deployments", deployment.plural_)
        self.assertEqual(True, Deployment.is_namespaced_)

    def test_defaults(self):
        self.assertNotEqual(None, K8sResource.patch_strategies_)

    def test_loading(self):
        secret_instance = Secret(
            metadata=ObjectMeta(name="some-secret", namespace="default"),
            data={"key": "value"}
        )
        self.assertEqual("v1", secret_instance.apiVersion)
        self.assertEqual("Secret", secret_instance.kind)

        dumped_secret = secret_instance.to_dict()
        self.assertEqual("v1", dumped_secret["apiVersion"])
        self.assertEqual("Secret", dumped_secret["kind"])

        loaded_secret = Secret.from_dict(dumped_secret)
        self.assertEqual(secret_instance, loaded_secret)

        copied_secret = copy.deepcopy(loaded_secret)
        self.assertEqual(secret_instance, copied_secret)

    def test_custom_resource_load(self):
        @dataclass(kw_only=True, frozen=True, slots=True)
        class NestedField(Loadable):
            value: str

        @dataclass(kw_only=True, frozen=True, slots=True)
        class MyCustomResourceSpec(Loadable):
            some_field: list[NestedField]

        @dataclass(kw_only=True, frozen=True, slots=True)
        class MyCustomResource(K8sResource):
            kind = "MyCustomResource"
            apiVersion = "my-api.com/v1alpha1"
            spec: MyCustomResourceSpec
            status: dict = field(default_factory=dict)
            is_namespaced_ = False
            group_ = "my-api.com"
            plural_ = "mycustomresources"

        resource_src = {
            "kind": "MyCustomResource",
            "apiVersion": "my-api.com/v1alpha1",
            "metadata": {"name": "test"},
            "spec": {"some_field": [{"value": "1"}, {"value": "2"}]}
        }
        res = MyCustomResource.from_dict(resource_src)
        self.assertEqual("2", res.spec.some_field[1].value)



# generated by https://github.com/puzl-cloud/kubesdk
#   filename:  apis__resource.k8s.io__v1_openapi.json

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import ClassVar, Dict, List, Set

from kube_models.const import *
from kube_models.loader import *
from kube_models.loader import Loadable
from kube_models.resource import *

from ...apimachinery.pkg.api.resource import Quantity
from ...apimachinery.pkg.apis.meta.v1 import Condition, ListMeta, ObjectMeta, Time
from ...apimachinery.pkg.runtime import RawExtension
from ..core.v1 import NodeSelector


@dataclass(slots=True, kw_only=True, frozen=True)
class CELDeviceSelector(Loadable):
    expression: str


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceAttribute(Loadable):
    bool: bool | None = None
    int: int | None = None
    string: str | None = None
    version: str | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceConstraint(Loadable):
    distinctAttribute: str | None = None
    matchAttribute: str | None = None
    requests: List[str] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceSelector(Loadable):
    cel: CELDeviceSelector | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceToleration(Loadable):
    effect: str | None = None
    key: str | None = None
    operator: str | None = 'Equal'
    tolerationSeconds: int | None = None
    value: str | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class NetworkDeviceData(Loadable):
    hardwareAddress: str | None = None
    interfaceName: str | None = None
    ips: List[str] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimConsumerReference(Loadable):
    name: str
    resource: str
    uid: str
    apiGroup: str | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourcePool(Loadable):
    generation: int
    name: str
    resourceSliceCount: int


@dataclass(slots=True, kw_only=True, frozen=True)
class CapacityRequestPolicyRange(Loadable):
    min: Quantity
    max: Quantity | None = None
    step: Quantity | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class CapacityRequirements(Loadable):
    requests: Dict[str, Quantity] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class Counter(Loadable):
    value: Quantity


@dataclass(slots=True, kw_only=True, frozen=True)
class CounterSet(Loadable):
    counters: Dict[str, Counter]
    name: str


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceCounterConsumption(Loadable):
    counterSet: str
    counters: Dict[str, Counter]


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceRequestAllocationResult(Loadable):
    device: str
    driver: str
    pool: str
    request: str
    adminAccess: bool | None = None
    bindingConditions: List[str] | None = None
    bindingFailureConditions: List[str] | None = None
    consumedCapacity: Dict[str, Quantity] | None = None
    shareID: str | None = None
    tolerations: List[DeviceToleration] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceSubRequest(Loadable):
    deviceClassName: str
    name: str
    allocationMode: str | None = None
    capacity: CapacityRequirements | None = None
    count: int | None = None
    selectors: List[DeviceSelector] | None = None
    tolerations: List[DeviceToleration] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceTaint(Loadable):
    effect: str
    key: str
    timeAdded: Time | None = None
    value: str | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class ExactDeviceRequest(Loadable):
    deviceClassName: str
    adminAccess: bool | None = None
    allocationMode: str | None = None
    capacity: CapacityRequirements | None = None
    count: int | None = None
    selectors: List[DeviceSelector] | None = None
    tolerations: List[DeviceToleration] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class OpaqueDeviceConfiguration(Loadable):
    driver: str
    parameters: RawExtension


@dataclass(slots=True, kw_only=True, frozen=True)
class AllocatedDeviceStatus(Loadable):
    device: str
    driver: str
    pool: str
    conditions: List[Condition] | None = None
    data: RawExtension | None = None
    networkData: NetworkDeviceData | None = None
    shareID: str | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class CapacityRequestPolicy(Loadable):
    default: Quantity | None = None
    validRange: CapacityRequestPolicyRange | None = None
    validValues: List[Quantity] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceAllocationConfiguration(Loadable):
    source: str
    opaque: OpaqueDeviceConfiguration | None = None
    requests: List[str] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceAllocationResult(Loadable):
    config: List[DeviceAllocationConfiguration] | None = None
    results: List[DeviceRequestAllocationResult] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceCapacity(Loadable):
    value: Quantity
    requestPolicy: CapacityRequestPolicy | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceClaimConfiguration(Loadable):
    opaque: OpaqueDeviceConfiguration | None = None
    requests: List[str] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceClassConfiguration(Loadable):
    opaque: OpaqueDeviceConfiguration | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceClassSpec(Loadable):
    config: List[DeviceClassConfiguration] | None = None
    extendedResourceName: str | None = None
    selectors: List[DeviceSelector] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceRequest(Loadable):
    name: str
    exactly: ExactDeviceRequest | None = None
    firstAvailable: List[DeviceSubRequest] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class AllocationResult(Loadable):
    allocationTimestamp: Time | None = None
    devices: DeviceAllocationResult | None = None
    nodeSelector: NodeSelector | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class Device(Loadable):
    name: str
    allNodes: bool | None = None
    allowMultipleAllocations: bool | None = None
    attributes: Dict[str, DeviceAttribute] | None = None
    bindingConditions: List[str] | None = None
    bindingFailureConditions: List[str] | None = None
    bindsToNode: bool | None = None
    capacity: Dict[str, DeviceCapacity] | None = None
    consumesCounters: List[DeviceCounterConsumption] | None = None
    nodeName: str | None = None
    nodeSelector: NodeSelector | None = None
    taints: List[DeviceTaint] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceClaim(Loadable):
    config: List[DeviceClaimConfiguration] | None = None
    constraints: List[DeviceConstraint] | None = None
    requests: List[DeviceRequest] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceClass(K8sResource):
    spec: DeviceClassSpec
    apiVersion: ClassVar[str] = 'resource.k8s.io/v1'
    kind: ClassVar[str] = 'DeviceClass'
    metadata: ObjectMeta = field(default_factory=ObjectMeta)
    plural_: ClassVar[str] = 'deviceclasses'
    is_namespaced_: ClassVar[bool] = False
    group_: ClassVar[Optional[str]] = 'resource.k8s.io'
    patch_strategies_: ClassVar[set[PatchRequestType]] = {
        'application/apply-patch+cbor',
        'application/apply-patch+yaml',
        'application/json-patch+json',
        'application/merge-patch+json',
        'application/strategic-merge-patch+json',
    }


@dataclass(slots=True, kw_only=True, frozen=True)
class DeviceClassList(Loadable):
    items: List[DeviceClass]
    apiVersion: str = 'resource.k8s.io/v1'
    kind: str = 'DeviceClassList'
    metadata: ListMeta = field(default_factory=ObjectMeta)


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimSpec(Loadable):
    devices: DeviceClaim | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimStatus(Loadable):
    allocation: AllocationResult | None = None
    devices: List[AllocatedDeviceStatus] | None = None
    reservedFor: List[ResourceClaimConsumerReference] = field(
        default_factory=list,
        metadata={
            'x-kubernetes-patch-strategy': 'merge',
            'x-kubernetes-patch-merge-key': 'uid',
        },
    )


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimTemplateSpec(Loadable):
    spec: ResourceClaimSpec
    metadata: ObjectMeta | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceSliceSpec(Loadable):
    driver: str
    pool: ResourcePool
    allNodes: bool | None = None
    devices: List[Device] | None = None
    nodeName: str | None = None
    nodeSelector: NodeSelector | None = None
    perDeviceNodeSelection: bool | None = None
    sharedCounters: List[CounterSet] | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaim(K8sResource):
    spec: ResourceClaimSpec
    apiVersion: ClassVar[str] = 'resource.k8s.io/v1'
    kind: ClassVar[str] = 'ResourceClaim'
    metadata: ObjectMeta = field(default_factory=ObjectMeta)
    status: ResourceClaimStatus | None = None
    plural_: ClassVar[str] = 'resourceclaims'
    is_namespaced_: ClassVar[bool] = True
    group_: ClassVar[Optional[str]] = 'resource.k8s.io'
    patch_strategies_: ClassVar[set[PatchRequestType]] = {
        'application/apply-patch+cbor',
        'application/apply-patch+yaml',
        'application/json-patch+json',
        'application/merge-patch+json',
        'application/strategic-merge-patch+json',
    }


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimList(Loadable):
    items: List[ResourceClaim]
    apiVersion: str = 'resource.k8s.io/v1'
    kind: str = 'ResourceClaimList'
    metadata: ListMeta = field(default_factory=ObjectMeta)


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimTemplate(K8sResource):
    spec: ResourceClaimTemplateSpec
    apiVersion: ClassVar[str] = 'resource.k8s.io/v1'
    kind: ClassVar[str] = 'ResourceClaimTemplate'
    metadata: ObjectMeta = field(default_factory=ObjectMeta)
    plural_: ClassVar[str] = 'resourceclaimtemplates'
    is_namespaced_: ClassVar[bool] = True
    group_: ClassVar[Optional[str]] = 'resource.k8s.io'
    patch_strategies_: ClassVar[set[PatchRequestType]] = {
        'application/apply-patch+cbor',
        'application/apply-patch+yaml',
        'application/json-patch+json',
        'application/merge-patch+json',
        'application/strategic-merge-patch+json',
    }


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceClaimTemplateList(Loadable):
    items: List[ResourceClaimTemplate]
    apiVersion: str = 'resource.k8s.io/v1'
    kind: str = 'ResourceClaimTemplateList'
    metadata: ListMeta = field(default_factory=ObjectMeta)


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceSlice(K8sResource):
    spec: ResourceSliceSpec
    apiVersion: ClassVar[str] = 'resource.k8s.io/v1'
    kind: ClassVar[str] = 'ResourceSlice'
    metadata: ObjectMeta = field(default_factory=ObjectMeta)
    plural_: ClassVar[str] = 'resourceslices'
    is_namespaced_: ClassVar[bool] = False
    group_: ClassVar[Optional[str]] = 'resource.k8s.io'
    patch_strategies_: ClassVar[set[PatchRequestType]] = {
        'application/apply-patch+cbor',
        'application/apply-patch+yaml',
        'application/json-patch+json',
        'application/merge-patch+json',
        'application/strategic-merge-patch+json',
    }


@dataclass(slots=True, kw_only=True, frozen=True)
class ResourceSliceList(Loadable):
    items: List[ResourceSlice]
    apiVersion: str = 'resource.k8s.io/v1'
    kind: str = 'ResourceSliceList'
    metadata: ListMeta = field(default_factory=ObjectMeta)
